@@ -34,9 +34,9 @@ from typing import Optional
 import click
 
 from dbt_osmosis.core.diff import diff_and_print_to_console
-from dbt_osmosis.core.logging import logger
-from dbt_osmosis.core.osmosis import (DEFAULT_PROFILES_DIR,
-                                      CompilationException, DbtOsmosis)
+from dbt_osmosis.core.log_controller import logger
+from dbt_osmosis.core.macros import inject_macros
+from dbt_osmosis.core.osmosis import DEFAULT_PROFILES_DIR, DbtOsmosis
 
 CONTEXT = {"max_content_width": 800}
 
@@ -312,23 +312,14 @@ def extract(
     default=DEFAULT_PROFILES_DIR,
     help="Which directory to look in for the profiles.yml file. Defaults to ~/.dbt",
 )
-@click.option(
-    "-m",
-    "--model",
-    type=click.STRING,
-    required=True,
-    help="The model to edit in the workbench",
-)
 @click.pass_context
-def workbench(
-    ctx, model: str, profiles_dir: Optional[str] = None, project_dir: Optional[str] = None
-):
+def workbench(ctx, profiles_dir: Optional[str] = None, project_dir: Optional[str] = None):
     """Instantiate the dbt-osmosis workbench and begin architecting a model.
         --model argument is required
 
     \f
     Pass the --options command to see streamlit specific options that can be passed to the app,
-
+    pass --config to see the output of streamlit config show
     """
 
     logger().info(":water_wave: Executing dbt-osmosis\n")
@@ -337,6 +328,16 @@ def workbench(
         subprocess.run(["streamlit", "run", "--help"])
         ctx.exit()
 
+    import os
+
+    if "--config" in ctx.args:
+        subprocess.run(
+            ["streamlit", "config", "show"],
+            env=os.environ,
+            cwd=Path.cwd(),
+        )
+        ctx.exit()
+
     script_args = ["--"]
     if project_dir:
         script_args.append("--project-dir")
@@ -344,14 +345,11 @@ def workbench(
     if profiles_dir:
         script_args.append("--profiles-dir")
         script_args.append(profiles_dir)
-    if model:
-        script_args.append("--model")
-        script_args.append(model)
-
-    import os
 
     subprocess.run(
-        ["streamlit", "run", Path(__file__).parent / "app.py"] + ctx.args + script_args,
+        ["streamlit", "run", "--runner.magicEnabled=false", Path(__file__).parent / "app.py"]
+        + ctx.args
+        + script_args,
         env=os.environ,
         cwd=Path.cwd(),
     )
@@ -380,13 +378,29 @@ def workbench(
     "--model",
     type=click.STRING,
     required=True,
-    help="The model to edit in the workbench",
+    help="The model to edit in the workbench, must be a valid model as would be selected by `ref`",
 )
 @click.option(
     "--pk",
     type=click.STRING,
     required=True,
-    help="The model to edit in the workbench",
+    help="The primary key of the model with which to base the diff",
+)
+@click.option(
+    "--temp-table",
+    is_flag=True,
+    help="If specified, temp tables are used to stage the queries.",
+)
+@click.option(
+    "--agg/--no-agg",
+    default=True,
+    help="Use --no-agg to show sample results, by default we agg for a summary view.",
+)
+@click.option(
+    "-o",
+    "--output",
+    default="table",
+    help="Output format can be one of table, chart/bar, or csv. CSV is saved to a file named dbt-osmosis-diff in working dir",
 )
 def diff(
     model: str,
@@ -394,6 +408,9 @@ def diff(
     target: Optional[str] = None,
     project_dir: Optional[str] = None,
     profiles_dir: Optional[str] = None,
+    temp_table: bool = False,
+    agg: bool = True,
+    output: str = "table",
 ):
     """Diff a model based on git HEAD to working copy on disk"""
 
@@ -404,7 +421,8 @@ def diff(
         profiles_dir=profiles_dir,
         target=target,
     )
-    diff_and_print_to_console(model, pk, runner)
+    inject_macros(runner)
+    diff_and_print_to_console(model, pk, runner, temp_table, agg, output)
 
 
 if __name__ == "__main__":
